@@ -10,125 +10,86 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import six
-
-from neutron.agent.ovsdb.native import commands
-from neutron.agent.ovsdb.native import idlutils
+from ovsdbapp.backend.ovs_idl import command
+from ovsdbapp.backend.ovs_idl import idlutils
 
 from networking_ovn._i18n import _
+from networking_ovn.common import constants as ovn_const
+from networking_ovn.common import exceptions as ovn_exc
 from networking_ovn.common import utils
 
-
-# TODO(rtheis): These wrapper functions are't needed once OpenStack
-# global requirements guarantee an ovs python version with mutate
-# support.
-
-def _is_ovs_mutate_available(row):
-    # Checking for the addvalue method should be sufficient.
-    return callable(getattr(row, 'addvalue', None))
+RESOURCE_TYPE_MAP = {
+    ovn_const.TYPE_NETWORKS: 'Logical_Switch',
+    ovn_const.TYPE_PORTS: 'Logical_Switch_Port',
+    ovn_const.TYPE_ROUTERS: 'Logical_Router',
+    ovn_const.TYPE_ROUTER_PORTS: 'Logical_Router_Port',
+    ovn_const.TYPE_FLOATINGIPS: 'NAT',
+    ovn_const.TYPE_SUBNETS: 'DHCP_Options',
+}
 
 
 def _addvalue_to_list(row, column, new_value):
-    # If available, use mutate support to add the value.
-    if _is_ovs_mutate_available(row):
-        row.addvalue(column, new_value)
-    else:
-        row.verify(column)
-        column_values = getattr(row, column, [])
-        if new_value not in column_values:
-            column_values.append(new_value)
-            setattr(row, column, column_values)
+    row.addvalue(column, new_value)
 
 
 def _delvalue_from_list(row, column, old_value):
-    # If available, use mutate support to delete the value.
-    if _is_ovs_mutate_available(row):
-        row.delvalue(column, old_value)
-    else:
-        row.verify(column)
-        column_values = getattr(row, column, [])
-        if old_value in column_values:
-            column_values.remove(old_value)
-            setattr(row, column, column_values)
+    row.delvalue(column, old_value)
 
 
 def _updatevalues_in_list(row, column, new_values=None, old_values=None):
     new_values = new_values or []
     old_values = old_values or []
 
-    # If available, use mutate support to add/delete the values.
-    if _is_ovs_mutate_available(row):
-        for new_value in new_values:
-            row.addvalue(column, new_value)
-        for old_value in old_values:
-            row.delvalue(column, old_value)
-    else:
-        row.verify(column)
-        column_values = getattr(row, column, [])
-        for new_value in new_values:
-            if new_value not in column_values:
-                column_values.append(new_value)
-        for old_value in old_values:
-            if old_value in column_values:
-                column_values.remove(old_value)
-        setattr(row, column, column_values)
+    for new_value in new_values:
+        row.addvalue(column, new_value)
+    for old_value in old_values:
+        row.delvalue(column, old_value)
 
 
-def get_lsp_dhcpv4_options_uuids(lsp, lsp_name):
-    # Get dhcpv4_options uuids from Logical_Switch_Port, which are references
-    # of port dhcpv4 options in DHCP_Options table.
+def get_lsp_dhcp_options_uuids(lsp, lsp_name):
+    # Get dhcpv4_options and dhcpv6_options uuids from Logical_Switch_Port,
+    # which are references of port dhcp options in DHCP_Options table.
     uuids = set()
     for dhcp_opts in getattr(lsp, 'dhcpv4_options', []):
         external_ids = getattr(dhcp_opts, 'external_ids', {})
         if external_ids.get('port_id') == lsp_name:
             uuids.add(dhcp_opts.uuid)
+    for dhcp_opts in getattr(lsp, 'dhcpv6_options', []):
+        external_ids = getattr(dhcp_opts, 'external_ids', {})
+        if external_ids.get('port_id') == lsp_name:
+            uuids.add(dhcp_opts.uuid)
     return uuids
 
 
-class AddLSwitchCommand(commands.BaseCommand):
-    def __init__(self, api, name, may_exist, **columns):
-        super(AddLSwitchCommand, self).__init__(api)
+def _add_gateway_chassis(api, txn, lrp_name, val):
+    gateway_chassis = api._tables.get('Gateway_Chassis')
+    if gateway_chassis:
+        prio = len(val)
+        uuid_list = []
+        for chassis in val:
+            gwc_name = '%s_%s' % (lrp_name, chassis)
+            try:
+                gwc = idlutils.row_by_value(api.idl,
+                                            'Gateway_Chassis',
+                                            'name', gwc_name)
+            except idlutils.RowNotFound:
+                gwc = txn.insert(gateway_chassis)
+                gwc.name = gwc_name
+            gwc.chassis_name = chassis
+            gwc.priority = prio
+            prio = prio - 1
+            uuid_list.append(gwc.uuid)
+        return 'gateway_chassis', uuid_list
+    else:
+        chassis = {ovn_const.OVN_GATEWAY_CHASSIS_KEY: val[0]}
+        return 'options', chassis
+
+
+class LSwitchSetExternalIdsCommand(command.BaseCommand):
+    def __init__(self, api, name, ext_ids, if_exists):
+        super(LSwitchSetExternalIdsCommand, self).__init__(api)
         self.name = name
-        self.columns = columns
-        self.may_exist = may_exist
-
-    def run_idl(self, txn):
-        if self.may_exist:
-            lswitch = idlutils.row_by_value(self.api.idl, 'Logical_Switch',
-                                            'name', self.name, None)
-            if lswitch:
-                return
-        row = txn.insert(self.api._tables['Logical_Switch'])
-        row.name = self.name
-        for col, val in self.columns.items():
-            setattr(row, col, val)
-
-
-class DelLSwitchCommand(commands.BaseCommand):
-    def __init__(self, api, name, if_exists):
-        super(DelLSwitchCommand, self).__init__(api)
-        self.name = name
-        self.if_exists = if_exists
-
-    def run_idl(self, txn):
-        try:
-            lswitch = idlutils.row_by_value(self.api.idl, 'Logical_Switch',
-                                            'name', self.name)
-        except idlutils.RowNotFound:
-            if self.if_exists:
-                return
-            msg = _("Logical Switch %s does not exist") % self.name
-            raise RuntimeError(msg)
-
-        self.api._tables['Logical_Switch'].rows[lswitch.uuid].delete()
-
-
-class LSwitchSetExternalIdCommand(commands.BaseCommand):
-    def __init__(self, api, name, field, value, if_exists):
-        super(LSwitchSetExternalIdCommand, self).__init__(api)
-        self.name = name
-        self.field = field
-        self.value = value
+        self.ext_ids = ext_ids
         self.if_exists = if_exists
 
     def run_idl(self, txn):
@@ -143,13 +104,13 @@ class LSwitchSetExternalIdCommand(commands.BaseCommand):
             raise RuntimeError(msg)
 
         lswitch.verify('external_ids')
-
         external_ids = getattr(lswitch, 'external_ids', {})
-        external_ids[self.field] = self.value
+        for key, value in self.ext_ids.items():
+            external_ids[key] = value
         lswitch.external_ids = external_ids
 
 
-class AddLSwitchPortCommand(commands.BaseCommand):
+class AddLSwitchPortCommand(command.BaseCommand):
     def __init__(self, api, lport, lswitch, may_exist, **columns):
         super(AddLSwitchPortCommand, self).__init__(api)
         self.lport = lport
@@ -173,13 +134,27 @@ class AddLSwitchPortCommand(commands.BaseCommand):
 
         port = txn.insert(self.api._tables['Logical_Switch_Port'])
         port.name = self.lport
+        dhcpv4_options = self.columns.pop('dhcpv4_options', [])
+        if isinstance(dhcpv4_options, list):
+            port.dhcpv4_options = dhcpv4_options
+        else:
+            port.dhcpv4_options = [dhcpv4_options.result]
+        dhcpv6_options = self.columns.pop('dhcpv6_options', [])
+        if isinstance(dhcpv6_options, list):
+            port.dhcpv6_options = dhcpv6_options
+        else:
+            port.dhcpv6_options = [dhcpv6_options.result]
         for col, val in self.columns.items():
             setattr(port, col, val)
         # add the newly created port to existing lswitch
         _addvalue_to_list(lswitch, 'ports', port.uuid)
+        self.result = port.uuid
+
+    def post_commit(self, txn):
+        self.result = txn.get_insert_uuid(self.result)
 
 
-class SetLSwitchPortCommand(commands.BaseCommand):
+class SetLSwitchPortCommand(command.BaseCommand):
     def __init__(self, api, lport, if_exists, **columns):
         super(SetLSwitchPortCommand, self).__init__(api)
         self.lport = lport
@@ -196,14 +171,34 @@ class SetLSwitchPortCommand(commands.BaseCommand):
             msg = _("Logical Switch Port %s does not exist") % self.lport
             raise RuntimeError(msg)
 
-        # Delete DHCP_Options records no longer refered by this port.
+        # Delete DHCP_Options records no longer referred by this port.
         # The table rows should be consistent for the same transaction.
-        # After we get a DHCP_Options row uuid from port dhcpv4_options
-        # reference, the row shouldn't disappear for this transaction,
-        # before we delete it.
-        cur_port_dhcp_opts = get_lsp_dhcpv4_options_uuids(
+        # After we get DHCP_Options rows uuids from port dhcpv4_options
+        # and dhcpv6_options references, the rows shouldn't disappear for
+        # this transaction before we delete it.
+        cur_port_dhcp_opts = get_lsp_dhcp_options_uuids(
             port, self.lport)
-        new_port_dhcp_opts = set(self.columns.get('dhcpv4_options', []))
+        new_port_dhcp_opts = set()
+        dhcpv4_options = self.columns.pop('dhcpv4_options', None)
+        if dhcpv4_options is None:
+            new_port_dhcp_opts.update([option.uuid for option in
+                                       getattr(port, 'dhcpv4_options', [])])
+        elif isinstance(dhcpv4_options, list):
+            new_port_dhcp_opts.update(dhcpv4_options)
+            port.dhcpv4_options = dhcpv4_options
+        else:
+            new_port_dhcp_opts.add(dhcpv4_options.result)
+            port.dhcpv4_options = [dhcpv4_options.result]
+        dhcpv6_options = self.columns.pop('dhcpv6_options', None)
+        if dhcpv6_options is None:
+            new_port_dhcp_opts.update([option.uuid for option in
+                                       getattr(port, 'dhcpv6_options', [])])
+        elif isinstance(dhcpv6_options, list):
+            new_port_dhcp_opts.update(dhcpv6_options)
+            port.dhcpv6_options = dhcpv6_options
+        else:
+            new_port_dhcp_opts.add(dhcpv6_options.result)
+            port.dhcpv6_options = [dhcpv6_options.result]
         for uuid in cur_port_dhcp_opts - new_port_dhcp_opts:
             self.api._tables['DHCP_Options'].rows[uuid].delete()
 
@@ -211,7 +206,7 @@ class SetLSwitchPortCommand(commands.BaseCommand):
             setattr(port, col, val)
 
 
-class DelLSwitchPortCommand(commands.BaseCommand):
+class DelLSwitchPortCommand(command.BaseCommand):
     def __init__(self, api, lport, lswitch, if_exists):
         super(DelLSwitchPortCommand, self).__init__(api)
         self.lport = lport
@@ -230,11 +225,17 @@ class DelLSwitchPortCommand(commands.BaseCommand):
             msg = _("Port %s does not exist") % self.lport
             raise RuntimeError(msg)
 
+        # Delete DHCP_Options records no longer referred by this port.
+        cur_port_dhcp_opts = get_lsp_dhcp_options_uuids(
+            lport, self.lport)
+        for uuid in cur_port_dhcp_opts:
+            self.api._tables['DHCP_Options'].rows[uuid].delete()
+
         _delvalue_from_list(lswitch, 'ports', lport)
         self.api._tables['Logical_Switch_Port'].rows[lport.uuid].delete()
 
 
-class AddLRouterCommand(commands.BaseCommand):
+class AddLRouterCommand(command.BaseCommand):
     def __init__(self, api, name, may_exist, **columns):
         super(AddLRouterCommand, self).__init__(api)
         self.name = name
@@ -254,7 +255,7 @@ class AddLRouterCommand(commands.BaseCommand):
             setattr(row, col, val)
 
 
-class UpdateLRouterCommand(commands.BaseCommand):
+class UpdateLRouterCommand(command.BaseCommand):
     def __init__(self, api, name, if_exists, **columns):
         super(UpdateLRouterCommand, self).__init__(api)
         self.name = name
@@ -277,7 +278,7 @@ class UpdateLRouterCommand(commands.BaseCommand):
             return
 
 
-class DelLRouterCommand(commands.BaseCommand):
+class DelLRouterCommand(command.BaseCommand):
     def __init__(self, api, name, if_exists):
         super(DelLRouterCommand, self).__init__(api)
         self.name = name
@@ -296,11 +297,12 @@ class DelLRouterCommand(commands.BaseCommand):
         self.api._tables['Logical_Router'].rows[lrouter.uuid].delete()
 
 
-class AddLRouterPortCommand(commands.BaseCommand):
-    def __init__(self, api, name, lrouter, **columns):
+class AddLRouterPortCommand(command.BaseCommand):
+    def __init__(self, api, name, lrouter, may_exist, **columns):
         super(AddLRouterPortCommand, self).__init__(api)
         self.name = name
         self.lrouter = lrouter
+        self.may_exist = may_exist
         self.columns = columns
 
     def run_idl(self, txn):
@@ -314,6 +316,8 @@ class AddLRouterPortCommand(commands.BaseCommand):
         try:
             idlutils.row_by_value(self.api.idl, 'Logical_Router_Port',
                                   'name', self.name)
+            if self.may_exist:
+                return
             # The LRP entry with certain name has already exist, raise an
             # exception to notice caller. It's caller's responsibility to
             # call UpdateLRouterPortCommand to get LRP entry processed
@@ -325,12 +329,15 @@ class AddLRouterPortCommand(commands.BaseCommand):
             lrouter_port = txn.insert(self.api._tables['Logical_Router_Port'])
             lrouter_port.name = self.name
             for col, val in self.columns.items():
+                if col == 'gateway_chassis':
+                    col, val = _add_gateway_chassis(self.api, txn, self.name,
+                                                    val)
                 setattr(lrouter_port, col, val)
             _addvalue_to_list(lrouter, 'ports', lrouter_port)
 
 
-class UpdateLRouterPortCommand(commands.BaseCommand):
-    def __init__(self, api, name, lrouter, if_exists, **columns):
+class UpdateLRouterPortCommand(command.BaseCommand):
+    def __init__(self, api, name, if_exists, **columns):
         super(UpdateLRouterPortCommand, self).__init__(api)
         self.name = name
         self.columns = columns
@@ -347,13 +354,21 @@ class UpdateLRouterPortCommand(commands.BaseCommand):
             msg = _("Logical Router Port %s does not exist") % self.name
             raise RuntimeError(msg)
 
-        if lrouter_port:
-            for col, val in self.columns.items():
-                setattr(lrouter_port, col, val)
-            return
+        # TODO(lucasagomes): Remove this check once we drop the support
+        # for OVS versions <= 2.8
+        ipv6_ra_configs_supported = self.api.is_col_present(
+            'Logical_Router_Port', 'ipv6_ra_configs')
+        for col, val in self.columns.items():
+            if col == 'ipv6_ra_configs' and not ipv6_ra_configs_supported:
+                continue
+
+            if col == 'gateway_chassis':
+                col, val = _add_gateway_chassis(self.api, txn, self.name,
+                                                val)
+            setattr(lrouter_port, col, val)
 
 
-class DelLRouterPortCommand(commands.BaseCommand):
+class DelLRouterPortCommand(command.BaseCommand):
     def __init__(self, api, name, lrouter, if_exists):
         super(DelLRouterPortCommand, self).__init__(api)
         self.name = name
@@ -380,27 +395,34 @@ class DelLRouterPortCommand(commands.BaseCommand):
         _delvalue_from_list(lrouter, 'ports', lrouter_port)
 
 
-class SetLRouterPortInLSwitchPortCommand(commands.BaseCommand):
-    def __init__(self, api, lswitch_port, lrouter_port):
+class SetLRouterPortInLSwitchPortCommand(command.BaseCommand):
+    def __init__(self, api, lswitch_port, lrouter_port, is_gw_port, if_exists):
         super(SetLRouterPortInLSwitchPortCommand, self).__init__(api)
         self.lswitch_port = lswitch_port
         self.lrouter_port = lrouter_port
+        self.is_gw_port = is_gw_port
+        self.if_exists = if_exists
 
     def run_idl(self, txn):
         try:
             port = idlutils.row_by_value(self.api.idl, 'Logical_Switch_Port',
                                          'name', self.lswitch_port)
         except idlutils.RowNotFound:
+            if self.if_exists:
+                return
             msg = _("Logical Switch Port %s does not "
                     "exist") % self.lswitch_port
             raise RuntimeError(msg)
 
         options = {'router-port': self.lrouter_port}
+        if self.is_gw_port:
+            options[ovn_const.OVN_GATEWAY_NAT_ADDRESSES_KEY] = 'router'
         setattr(port, 'options', options)
         setattr(port, 'type', 'router')
+        setattr(port, 'addresses', 'router')
 
 
-class AddACLCommand(commands.BaseCommand):
+class AddACLCommand(command.BaseCommand):
     def __init__(self, api, lswitch, lport, **columns):
         super(AddACLCommand, self).__init__(api)
         self.lswitch = lswitch
@@ -418,11 +440,10 @@ class AddACLCommand(commands.BaseCommand):
         row = txn.insert(self.api._tables['ACL'])
         for col, val in self.columns.items():
             setattr(row, col, val)
-        row.external_ids = {'neutron:lport': self.lport}
         _addvalue_to_list(lswitch, 'acls', row.uuid)
 
 
-class DelACLCommand(commands.BaseCommand):
+class DelACLCommand(command.BaseCommand):
     def __init__(self, api, lswitch, lport, if_exists):
         super(DelACLCommand, self).__init__(api)
         self.lswitch = lswitch
@@ -450,7 +471,7 @@ class DelACLCommand(commands.BaseCommand):
         _updatevalues_in_list(lswitch, 'acls', old_values=acls_to_del)
 
 
-class UpdateACLsCommand(commands.BaseCommand):
+class UpdateACLsCommand(command.BaseCommand):
     def __init__(self, api, lswitch_names, port_list, acl_new_values_dict,
                  need_compare=True, is_add_acl=True):
         """This command updates the acl list for the logical switches
@@ -548,15 +569,18 @@ class UpdateACLsCommand(commands.BaseCommand):
         else:
             acl_add_values_dict = {}
             acl_del_objs_dict = {}
-            del_acl_matches = []
+            del_acl_extids = []
             for acl_dict in self.acl_new_values_dict.values():
-                del_acl_matches.append(acl_dict['match'])
-            for switch_name, lswitch in six.iteritems(lswitch_ovsdb_dict):
+                del_acl_extids.append({acl_dict['match']:
+                                       acl_dict['external_ids']})
+            for switch_name, lswitch in lswitch_ovsdb_dict.items():
                 if switch_name not in acl_del_objs_dict:
                     acl_del_objs_dict[switch_name] = []
                 acls = getattr(lswitch, 'acls', [])
                 for acl in acls:
-                    if getattr(acl, 'match') in del_acl_matches:
+                    match = getattr(acl, 'match')
+                    acl_extids = {match: getattr(acl, 'external_ids')}
+                    if acl_extids in del_acl_extids:
                         acl_del_objs_dict[switch_name].append(acl)
         return lswitch_ovsdb_dict, acl_del_objs_dict, acl_add_values_dict
 
@@ -576,7 +600,7 @@ class UpdateACLsCommand(commands.BaseCommand):
             lswitch_ovsdb_dict, acl_del_objs_dict, acl_add_values_dict = \
                 self._get_update_data_without_compare()
 
-        for lswitch_name, lswitch in six.iteritems(lswitch_ovsdb_dict):
+        for lswitch_name, lswitch in lswitch_ovsdb_dict.items():
             acl_del_objs = acl_del_objs_dict.get(lswitch_name, [])
             acl_add_values = acl_add_values_dict.get(lswitch_name, [])
 
@@ -605,7 +629,7 @@ class UpdateACLsCommand(commands.BaseCommand):
                                   old_values=acl_del_objs)
 
 
-class AddStaticRouteCommand(commands.BaseCommand):
+class AddStaticRouteCommand(command.BaseCommand):
     def __init__(self, api, lrouter, **columns):
         super(AddStaticRouteCommand, self).__init__(api)
         self.lrouter = lrouter
@@ -625,7 +649,7 @@ class AddStaticRouteCommand(commands.BaseCommand):
         _addvalue_to_list(lrouter, 'static_routes', row.uuid)
 
 
-class DelStaticRouteCommand(commands.BaseCommand):
+class DelStaticRouteCommand(command.BaseCommand):
     def __init__(self, api, lrouter, ip_prefix, nexthop, if_exists):
         super(DelStaticRouteCommand, self).__init__(api)
         self.lrouter = lrouter
@@ -653,7 +677,7 @@ class DelStaticRouteCommand(commands.BaseCommand):
                 break
 
 
-class AddAddrSetCommand(commands.BaseCommand):
+class AddAddrSetCommand(command.BaseCommand):
     def __init__(self, api, name, may_exist, **columns):
         super(AddAddrSetCommand, self).__init__(api)
         self.name = name
@@ -672,7 +696,7 @@ class AddAddrSetCommand(commands.BaseCommand):
             setattr(row, col, val)
 
 
-class DelAddrSetCommand(commands.BaseCommand):
+class DelAddrSetCommand(command.BaseCommand):
     def __init__(self, api, name, if_exists):
         super(DelAddrSetCommand, self).__init__(api)
         self.name = name
@@ -692,7 +716,7 @@ class DelAddrSetCommand(commands.BaseCommand):
         self.api._tables['Address_Set'].rows[addrset.uuid].delete()
 
 
-class UpdateAddrSetCommand(commands.BaseCommand):
+class UpdateAddrSetCommand(command.BaseCommand):
     def __init__(self, api, name, addrs_add, addrs_remove, if_exists):
         super(UpdateAddrSetCommand, self).__init__(api)
         self.name = name
@@ -717,7 +741,7 @@ class UpdateAddrSetCommand(commands.BaseCommand):
             old_values=self.addrs_remove)
 
 
-class UpdateAddrSetExtIdsCommand(commands.BaseCommand):
+class UpdateAddrSetExtIdsCommand(command.BaseCommand):
     def __init__(self, api, name, external_ids, if_exists):
         super(UpdateAddrSetExtIdsCommand, self).__init__(api)
         self.name = name
@@ -737,19 +761,70 @@ class UpdateAddrSetExtIdsCommand(commands.BaseCommand):
 
         addrset.verify('external_ids')
         addrset_external_ids = getattr(addrset, 'external_ids', {})
-        for ext_id_key, ext_id_value in six.iteritems(self.external_ids):
+        for ext_id_key, ext_id_value in self.external_ids.items():
             addrset_external_ids[ext_id_key] = ext_id_value
         addrset.external_ids = addrset_external_ids
 
 
-class AddDHCPOptionsCommand(commands.BaseCommand):
-    def __init__(self, api, subnet_id, port_id=None, may_exists=True,
+class UpdateChassisExtIdsCommand(command.BaseCommand):
+    def __init__(self, api, name, external_ids, if_exists):
+        super(UpdateChassisExtIdsCommand, self).__init__(api)
+        self.name = name
+        self.external_ids = external_ids
+        self.if_exists = if_exists
+
+    def run_idl(self, txn):
+        try:
+            chassis = idlutils.row_by_value(self.api.idl, 'Chassis',
+                                            'name', self.name)
+        except idlutils.RowNotFound:
+            if self.if_exists:
+                return
+            msg = _("Chassis %s does not exist. "
+                    "Can't update external IDs") % self.name
+            raise RuntimeError(msg)
+
+        chassis.verify('external_ids')
+        chassis_external_ids = getattr(chassis, 'external_ids', {})
+        for ext_id_key, ext_id_value in self.external_ids.items():
+            chassis_external_ids[ext_id_key] = ext_id_value
+        chassis.external_ids = chassis_external_ids
+
+
+class UpdatePortBindingExtIdsCommand(command.BaseCommand):
+    def __init__(self, api, name, external_ids, if_exists):
+        super(UpdatePortBindingExtIdsCommand, self).__init__(api)
+        self.name = name
+        self.external_ids = external_ids
+        self.if_exists = if_exists
+
+    def run_idl(self, txn):
+        try:
+            port = idlutils.row_by_value(self.api.idl, 'Port_Binding',
+                                         'logical_port', self.name)
+        except idlutils.RowNotFound:
+            if self.if_exists:
+                return
+            msg = _("Port %s does not exist. "
+                    "Can't update external IDs") % self.name
+            raise RuntimeError(msg)
+
+        port.verify('external_ids')
+        port_external_ids = getattr(port, 'external_ids', {})
+        for ext_id_key, ext_id_value in self.external_ids.items():
+            port_external_ids[ext_id_key] = ext_id_value
+        port.external_ids = port_external_ids
+
+
+class AddDHCPOptionsCommand(command.BaseCommand):
+    def __init__(self, api, subnet_id, port_id=None, may_exist=True,
                  **columns):
         super(AddDHCPOptionsCommand, self).__init__(api)
         self.columns = columns
-        self.may_exists = may_exists
+        self.may_exist = may_exist
         self.subnet_id = subnet_id
         self.port_id = port_id
+        self.new_insert = False
 
     def _get_dhcp_options_row(self):
         for row in self.api._tables['DHCP_Options'].rows.values():
@@ -761,16 +836,24 @@ class AddDHCPOptionsCommand(commands.BaseCommand):
 
     def run_idl(self, txn):
         row = None
-        if self.may_exists:
+        if self.may_exist:
             row = self._get_dhcp_options_row()
 
         if not row:
             row = txn.insert(self.api._tables['DHCP_Options'])
+            self.new_insert = True
         for col, val in self.columns.items():
             setattr(row, col, val)
+        self.result = row.uuid
+
+    def post_commit(self, txn):
+        # Update the result with inserted uuid for new inserted row, or the
+        # uuid get in run_idl should be real uuid already.
+        if self.new_insert:
+            self.result = txn.get_insert_uuid(self.result)
 
 
-class DelDHCPOptionsCommand(commands.BaseCommand):
+class DelDHCPOptionsCommand(command.BaseCommand):
     def __init__(self, api, row_uuid, if_exists=True):
         super(DelDHCPOptionsCommand, self).__init__(api)
         self.if_exists = if_exists
@@ -784,3 +867,287 @@ class DelDHCPOptionsCommand(commands.BaseCommand):
             raise RuntimeError(msg)
 
         self.api._tables['DHCP_Options'].rows[self.row_uuid].delete()
+
+
+class AddNATRuleInLRouterCommand(command.BaseCommand):
+    # TODO(chandrav): Add unit tests, bug #1638715.
+    def __init__(self, api, lrouter, **columns):
+        super(AddNATRuleInLRouterCommand, self).__init__(api)
+        self.lrouter = lrouter
+        self.columns = columns
+
+    def run_idl(self, txn):
+        try:
+            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
+                                            'name', self.lrouter)
+        except idlutils.RowNotFound:
+            msg = _("Logical Router %s does not exist") % self.lrouter
+            raise RuntimeError(msg)
+
+        row = txn.insert(self.api._tables['NAT'])
+        for col, val in self.columns.items():
+            setattr(row, col, val)
+        # TODO(chandrav): convert this to ovs transaction mutate
+        lrouter.verify('nat')
+        nat = getattr(lrouter, 'nat', [])
+        nat.append(row.uuid)
+        setattr(lrouter, 'nat', nat)
+
+
+class DeleteNATRuleInLRouterCommand(command.BaseCommand):
+    # TODO(chandrav): Add unit tests, bug #1638715.
+    def __init__(self, api, lrouter, type, logical_ip, external_ip,
+                 if_exists):
+        super(DeleteNATRuleInLRouterCommand, self).__init__(api)
+        self.lrouter = lrouter
+        self.type = type
+        self.logical_ip = logical_ip
+        self.external_ip = external_ip
+        self.if_exists = if_exists
+
+    def run_idl(self, txn):
+        try:
+            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
+                                            'name', self.lrouter)
+        except idlutils.RowNotFound:
+            if self.if_exists:
+                return
+            msg = _("Logical Router %s does not exist") % self.lrouter
+            raise RuntimeError(msg)
+
+        lrouter.verify('nat')
+        # TODO(chandrav): convert this to ovs transaction mutate
+        nats = getattr(lrouter, 'nat', [])
+        for nat in nats:
+            type = getattr(nat, 'type', '')
+            external_ip = getattr(nat, 'external_ip', '')
+            logical_ip = getattr(nat, 'logical_ip', '')
+            if self.type == type and \
+               self.external_ip == external_ip and \
+               self.logical_ip == logical_ip:
+                nats.remove(nat)
+                nat.delete()
+                break
+        setattr(lrouter, 'nat', nats)
+
+
+class SetNATRuleInLRouterCommand(command.BaseCommand):
+    def __init__(self, api, lrouter, nat_rule_uuid, **columns):
+        super(SetNATRuleInLRouterCommand, self).__init__(api)
+        self.lrouter = lrouter
+        self.nat_rule_uuid = nat_rule_uuid
+        self.columns = columns
+
+    def run_idl(self, txn):
+        try:
+            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
+                                            'name', self.lrouter)
+        except idlutils.RowNotFound:
+            msg = _("Logical Router %s does not exist") % self.lrouter
+            raise RuntimeError(msg)
+
+        lrouter.verify('nat')
+        nat_rules = getattr(lrouter, 'nat', [])
+        for nat_rule in nat_rules:
+            if nat_rule.uuid == self.nat_rule_uuid:
+                for col, val in self.columns.items():
+                    setattr(nat_rule, col, val)
+                break
+
+
+class AddNatIpToLRPortPeerOptionsCommand(command.BaseCommand):
+    # TODO(chandrav): Add unit tests, bug #1638715.
+    def __init__(self, api, lport, nat_ip):
+        super(AddNatIpToLRPortPeerOptionsCommand, self).__init__(api)
+        self.lport = lport
+        self.nat_ip = nat_ip
+
+    def run_idl(self, txn):
+        try:
+            lport = idlutils.row_by_value(self.api.idl, 'Logical_Switch_Port',
+                                          'name', self.lport)
+        except idlutils.RowNotFound:
+            msg = _("Logical Switch Port %s does not exist") % self.lport
+            raise RuntimeError(msg)
+
+        lport.verify('addresses')
+        addresses = getattr(lport, 'addresses', [])
+        lport.verify('options')
+        options = getattr(lport, 'options', {})
+
+        nat_ip_list = options.get('nat-addresses', '').split()
+        if not nat_ip_list:
+            nat_ips = addresses[0].split()[0] + ' ' + self.nat_ip
+        elif self.nat_ip not in nat_ip_list:
+            nat_ip_list.append(self.nat_ip)
+            nat_ips = ' '.join(nat_ip_list)
+        else:
+            return
+        options['nat-addresses'] = nat_ips
+        lport.options = options
+
+
+class DeleteNatIpFromLRPortPeerOptionsCommand(command.BaseCommand):
+    # TODO(chandrav): Add unit tests, bug #1638715.
+    def __init__(self, api, lport, nat_ip):
+        super(DeleteNatIpFromLRPortPeerOptionsCommand, self).__init__(api)
+        self.lport = lport
+        self.nat_ip = nat_ip
+
+    def run_idl(self, txn):
+        try:
+            lport = idlutils.row_by_value(self.api.idl, 'Logical_Switch_Port',
+                                          'name', self.lport)
+
+        except idlutils.RowNotFound:
+            msg = _("Logical Switch Port %s does not exist") % self.lport
+            raise RuntimeError(msg)
+
+        lport.verify('options')
+        options = getattr(lport, 'options', {})
+        nat_ip_list = options.get('nat-addresses', '').split()
+        if self.nat_ip not in nat_ip_list:
+            return
+        nat_ip_list.remove(self.nat_ip)
+        if len(nat_ip_list) is 1:
+            nat_ips = ''
+        else:
+            nat_ips = ' '.join(nat_ip_list)
+        if not nat_ips:
+            del options['nat-addresses']
+        else:
+            options['nat-addresses'] = nat_ips
+        lport.options = options
+
+
+class CheckRevisionNumberCommand(command.BaseCommand):
+
+    def __init__(self, api, name, resource, resource_type, if_exists):
+        super(CheckRevisionNumberCommand, self).__init__(api)
+        self.name = name
+        self.resource = resource
+        self.resource_type = resource_type
+        self.if_exists = if_exists
+
+    def _get_floatingip(self):
+        # TODO(lucasagomes): We can't use self.api.lookup() because that
+        # method does not introspect map type columns. We could either:
+        # 1. Enhance it to look into maps or, 2. Add a new ``name`` column
+        # to the NAT table so that we can use lookup() just like we do
+        # for other resources
+        for nat in self.api._tables['NAT'].rows.values():
+            if nat.type != 'dnat_and_snat':
+                continue
+            ext_ids = getattr(nat, 'external_ids', {})
+            if ext_ids.get(ovn_const.OVN_FIP_EXT_ID_KEY) == self.name:
+                return nat
+
+        raise idlutils.RowNotFound(
+            table='NAT', col='external_ids', match=self.name)
+
+    def _get_subnet(self):
+        for dhcp in self.api._tables['DHCP_Options'].rows.values():
+            ext_ids = getattr(dhcp, 'external_ids', {})
+            # Ignore ports DHCP Options
+            if ext_ids.get('port_id'):
+                continue
+            if ext_ids.get('subnet_id') == self.name:
+                return dhcp
+
+        raise idlutils.RowNotFound(
+            table='DHCP_Options', col='external_ids', match=self.name)
+
+    def run_idl(self, txn):
+        try:
+            ovn_table = RESOURCE_TYPE_MAP[self.resource_type]
+            # TODO(lucasagomes): After OVS 2.8.2 is released all tables should
+            # have the external_ids column. We can remove this conditional
+            # here by then.
+            if not self.api.is_col_present(ovn_table, 'external_ids'):
+                return
+
+            ovn_resource = None
+            if self.resource_type == ovn_const.TYPE_FLOATINGIPS:
+                ovn_resource = self._get_floatingip()
+            elif self.resource_type == ovn_const.TYPE_SUBNETS:
+                ovn_resource = self._get_subnet()
+            else:
+                ovn_resource = self.api.lookup(ovn_table, self.name)
+        except idlutils.RowNotFound:
+            if self.if_exists:
+                return
+            msg = (_('Failed to check the revision number for %s: Resource '
+                     'does not exist') % self.name)
+            raise RuntimeError(msg)
+
+        external_ids = getattr(ovn_resource, 'external_ids', {})
+        ovn_revision = int(external_ids.get(
+            ovn_const.OVN_REV_NUM_EXT_ID_KEY, -1))
+        neutron_revision = utils.get_revision_number(self.resource,
+                                                     self.resource_type)
+        if ovn_revision > neutron_revision:
+            raise ovn_exc.RevisionConflict(
+                resource_id=self.name, resource_type=self.resource_type)
+
+        ovn_resource.verify('external_ids')
+        ovn_resource.setkey('external_ids', ovn_const.OVN_REV_NUM_EXT_ID_KEY,
+                            str(neutron_revision))
+
+    def post_commit(self, txn):
+        self.result = ovn_const.TXN_COMMITTED
+
+
+class DeleteLRouterExtGwCommand(command.BaseCommand):
+
+    def __init__(self, api, lrouter, if_exists):
+        super(DeleteLRouterExtGwCommand, self).__init__(api)
+        self.lrouter = lrouter
+        self.if_exists = if_exists
+
+    def run_idl(self, txn):
+        # TODO(lucasagomes): Remove this check after OVS 2.8.2 is tagged
+        # (prior to that, the external_ids column didn't exist in this
+        # table).
+        if not self.api.is_col_present('Logical_Router_Static_Route',
+                                       'external_ids'):
+            return
+
+        try:
+            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
+                                            'name', self.lrouter)
+        except idlutils.RowNotFound:
+            if self.if_exists:
+                return
+            msg = _("Logical Router %s does not exist") % self.lrouter
+            raise RuntimeError(msg)
+
+        lrouter.verify('static_routes')
+        static_routes = getattr(lrouter, 'static_routes', [])
+        for route in static_routes:
+            external_ids = getattr(route, 'external_ids', {})
+            if ovn_const.OVN_ROUTER_IS_EXT_GW in external_ids:
+                _delvalue_from_list(lrouter, 'static_routes', route)
+                route.delete()
+                break
+
+        lrouter.verify('nat')
+        nats = getattr(lrouter, 'nat', [])
+        for nat in nats:
+            if nat.type != 'snat':
+                continue
+            _delvalue_from_list(lrouter, 'nat', nat)
+            nat.delete()
+
+        lrouter_ext_ids = getattr(lrouter, 'external_ids', {})
+        gw_port_id = lrouter_ext_ids.get(ovn_const.OVN_GW_PORT_EXT_ID_KEY)
+        if not gw_port_id:
+            return
+
+        try:
+            lrouter_port = idlutils.row_by_value(
+                self.api.idl, 'Logical_Router_Port', 'name',
+                utils.ovn_lrouter_port_name(gw_port_id))
+        except idlutils.RowNotFound:
+            return
+
+        _delvalue_from_list(lrouter, 'ports', lrouter_port)
