@@ -11,64 +11,27 @@
 #    under the License.
 
 import abc
+
+from ovsdbapp import api
 import six
 
 
 @six.add_metaclass(abc.ABCMeta)
-class API(object):
+class API(api.API):
 
     @abc.abstractmethod
-    def transaction(self, check_error=False, log_errors=True, **kwargs):
-        """Create a transaction
-
-        :param check_error: Allow the transaction to raise an exception?
-        :type check_error:  bool
-        :param log_errors:  Log an error if the transaction fails?
-        :type log_errors:   bool
-        :returns: A new transaction
-        :rtype: :class:`Transaction`
-        """
-
-    @abc.abstractmethod
-    def create_lswitch(self, name, may_exist=True, **columns):
-        """Create a command to add an OVN lswitch
-
-        :param name:         The id of the lswitch
-        :type name:          string
-        :param may_exist:    Do not fail if lswitch already exists
-        :type may_exist:     bool
-        :param columns:      Dictionary of lswitch columns
-                             Supported columns: external_ids
-        :type columns:       dictionary
-        :returns:            :class:`Command` with no result
-        """
-
-    @abc.abstractmethod
-    def set_lswitch_ext_id(self, name, ext_id, if_exists=True):
-        """Create a command to set OVN lswitch external id
+    def set_lswitch_ext_ids(self, name, ext_ids, if_exists=True):
+        """Create a command to set OVN lswitch external ids
 
         :param name:      The name of the lswitch
         :type name:       string
-        :param ext_id:    The external id to set for the lswitch
-        :type ext_id:     pair of <ext_id_key ,ext_id_value>
+        :param ext_ids    The external ids to set for the lswitch
+        :type ext_ids:    dictionary
         :param if_exists: Do not fail if lswitch does not exist
         :type if_exists:  bool
         :returns:        :class:`Command` with no result
         """
 
-    @abc.abstractmethod
-    def delete_lswitch(self, name=None, ext_id=None, if_exists=True):
-        """Create a command to delete an OVN lswitch
-
-        :param name:      The name of the lswitch
-        :type name:       string
-        :param ext_id:    The external id of the lswitch
-        :type ext_id:     pair of <ext_id_key ,ext_id_value>
-        :param if_exists: Do not fail if the lswitch does not exists
-        :type if_exists:  bool
-        :returns:         :class:`Command` with no result
-        """
-
     @abc.abstractmethod
     def create_lswitch_port(self, lport_name, lswitch_name, may_exist=True,
                             **columns):
@@ -113,32 +76,11 @@ class API(object):
         :type lswitch_name:   string
         :param ext_id:        The external id of the lport
         :type ext_id:         pair of <ext_id_key ,ext_id_value>
-        :param if_exists:     Do not fail if the lport does not exists
+        :param if_exists:     Do not fail if the lport does not exist
         :type if_exists:      bool
         :returns:             :class:`Command` with no result
         """
 
-    @abc.abstractmethod
-    def get_all_logical_switches_ids(self):
-        """Returns all logical switches names and external ids
-
-        :returns: dictionary with lswitch name and ext ids
-        """
-
-    @abc.abstractmethod
-    def get_logical_switch_ids(self, lswitch_name):
-        """Get external_ids for a Logical_Switch.
-
-        :returns: dict of external_ids.
-        """
-
-    @abc.abstractmethod
-    def get_all_logical_switch_ports_ids(self):
-        """Returns all logical switch ports names and external ids
-
-        :returns: dictionary with lsp name and ext ids
-        """
-
     @abc.abstractmethod
     def create_lrouter(self, name, may_exist=True, **columns):
         """Create a command to add an OVN lrouter
@@ -159,7 +101,7 @@ class API(object):
 
         :param name:         The id of the lrouter
         :type name:          string
-        :param if_exists:    Do not fail if the lrouter  does not exists
+        :param if_exists:    Do not fail if the lrouter does not exist
         :type if_exists:     bool
         :param columns:      Dictionary of lrouter columns
                              Supported columns: external_ids, default_gw, ip
@@ -173,13 +115,13 @@ class API(object):
 
         :param name:         The id of the lrouter
         :type name:          string
-        :param if_exists:    Do not fail if the lrouter  does not exists
+        :param if_exists:    Do not fail if the lrouter does not exist
         :type if_exists:     bool
         :returns:            :class:`Command` with no result
         """
 
     @abc.abstractmethod
-    def add_lrouter_port(self, name, lrouter, if_exists=True,
+    def add_lrouter_port(self, name, lrouter, may_exist=True,
                          **columns):
         """Create a command to add an OVN lrouter port
 
@@ -189,8 +131,9 @@ class API(object):
         :type lrouter:       string
         :param lswitch:      The unique name of the lswitch
         :type lswitch:       string
-        :param if_exists:    Do not fail if lrouter port already exists
-        :type if_exists:     bool
+        :param may_exist:    If true, do not fail if lrouter port set
+                             already exists.
+        :type may_exist:     bool
         :param columns:      Dictionary of lrouter columns
                              Supported columns: external_ids, mac, network
         :type columns:       dictionary
@@ -198,14 +141,12 @@ class API(object):
         """
 
     @abc.abstractmethod
-    def update_lrouter_port(self, name, lrouter, if_exists=True, **columns):
+    def update_lrouter_port(self, name, if_exists=True, **columns):
         """Update a command to add an OVN lrouter port
 
         :param name:         The unique name of the lrouter port
         :type name:          string
-        :param lrouter:      The unique name of the lrouter
-        :type lrouter:       string
-        :param if_exists:    Do not fail if the lrouter port does not exists
+        :param if_exists:    Do not fail if the lrouter port does not exist
         :type if_exists:     bool
         :param columns:      Dictionary of lrouter columns
                              Supported columns: networks
@@ -221,19 +162,24 @@ class API(object):
         :type name:          string
         :param lrouter:      The unique name of the lrouter
         :type lrouter:       string
-        :param if_exists:    Do not fail if the lrouter port does not exists
+        :param if_exists:    Do not fail if the lrouter port does not exist
         :type if_exists:     bool
         :returns:            :class:`Command` with no result
         """
 
     @abc.abstractmethod
-    def set_lrouter_port_in_lswitch_port(self, lswitch_port, lrouter_port):
+    def set_lrouter_port_in_lswitch_port(self, lswitch_port, lrouter_port,
+                                         is_gw_port=False, if_exists=True):
         """Create a command to set lswitch_port as lrouter_port
 
         :param lswitch_port: The name of logical switch port
         :type lswitch_port:  string
         :param lrouter_port: The name of logical router port
         :type lrouter_port:  string
+        :param is_gw_port:   True if logical router port is gw port
+        :type is_gw_port:    bool
+        :param if_exists:    Do not fail if the lswitch port does not exist
+        :type if_exists:     bool
         :returns:            :class:`Command` with no result
         """
 
@@ -281,6 +227,15 @@ class API(object):
         :type is_add_acl:             bool
         """
 
+    @abc.abstractmethod
+    def get_acl_by_id(self, acl_id):
+        """Get an ACL by its ID.
+
+        :param acl_id:                ID of the ACL to lookup
+        :type acl_id:                 string
+        :returns                      The ACL row or None:
+        """
+
     @abc.abstractmethod
     def add_static_route(self, lrouter, **columns):
         """Add static route to logical router.
@@ -363,8 +318,9 @@ class API(object):
         """
 
     @abc.abstractmethod
-    def get_all_chassis_router_bindings(self, chassis_candidate_list=None):
-        """Return a dictionary of chassis name:list of router gateways
+    def get_all_chassis_gateway_bindings(self,
+                                         chassis_candidate_list=None):
+        """Return a dictionary of chassis name:list of gateways
 
         :param chassis_candidate_list:  List of possible chassis candidates
         :type chassis_candidate_list:   []
@@ -372,24 +328,35 @@ class API(object):
         """
 
     @abc.abstractmethod
-    def get_router_chassis_binding(self, router_id):
-        """Return the chassis to which the router is bound to
+    def get_gateway_chassis_binding(self, gateway_id):
+        """Return the list of chassis to which the gateway is bound to
 
-        :param router_id:     The neutron router id
-        :type router_id:      string
-        :returns              string containing the chassis name
+        As one gateway can be hosted by multiple chassis, this method is
+        returning a list of those chassis ordered by priority. This means
+        that the first element of the list is the chassis hosting the
+        gateway with the highest priority (which will likely be where
+        the router port is going to be active).
+
+        :param gateway_id:     The gateway id
+        :type gateway_id:      string
+        :returns:              a list of strings with the chassis names
         """
 
     @abc.abstractmethod
-    def get_unhosted_routers(self, valid_chassis_list):
-        """Return a dictionary of routers gateways not hosted on chassis
+    def get_unhosted_gateways(self, port_physnet_dict, chassis_physnets,
+                              gw_chassis):
+        """Return a list of gateways not hosted on chassis
 
-        :param valid_chassis_list: List of valid chassis
-        :returns                   List of routers not hosted on a valid
-                                   chassis
+        :param port_physnet_dict: Dictionary of gateway ports and their physnet
+        :param chassis_physnets:  Dictionary of chassis and physnets
+        :param gw_chassis:        List of gateway chassis provided by admin
+                                  through ovn-cms-options
+        :returns:                 List of gateways not hosted on a valid
+                                  chassis
         """
 
-    def add_dhcp_options(self, subnet_id, port_id=None, may_exists=True,
+    @abc.abstractmethod
+    def add_dhcp_options(self, subnet_id, port_id=None, may_exist=True,
                          **columns):
         """Adds the DHCP options specified in the @columns in DHCP_Options
 
@@ -403,11 +370,11 @@ class API(object):
         :param port_id:        The port id to which the DHCP options belong to
                                if specified
         :type port_id:         string
-        :param may_exists:     If true, checks if the DHCP options for
+        :param may_exist:      If true, checks if the DHCP options for
                                subnet_id exists or not. If it already exists,
                                it updates the row with the columns specified.
                                Else creates a new row.
-        :type may_exists:      bool
+        :type may_exist:       bool
         :type columns:         Dictionary of DHCP_Options columns
                                Supported columns: see DHCP_Options table in
                                OVN_Northbound
@@ -426,42 +393,26 @@ class API(object):
         """
 
     @abc.abstractmethod
-    def get_subnet_dhcp_options(self, subnet_id):
+    def get_subnet_dhcp_options(self, subnet_id, with_ports=False):
         """Returns the Subnet DHCP options as a dictionary
 
         :param subnet_id:      The subnet id whose DHCP options are returned
         :type subnet_id:       string
-        :returns:              Returns the columns of the DHCP_Options as a
-                               dictionary. None is returned if no DHCP options.
+        :param with_ports:     If True, also returns the ports DHCP options.
+        :type with_ports:      bool
+        :returns:              Returns a dictionary containing two keys:
+                               subnet and ports.
         """
 
     @abc.abstractmethod
-    def get_port_dhcp_options(self, subnet_id, port_id):
-        """Returns the Logical switch port DHCP_Options as a dictionary
+    def get_subnets_dhcp_options(self, subnet_ids):
+        """Returns the Subnets DHCP options as list of dictionary
 
-        :param subnet_id:      The subnet id whose DHCP options are returned
-        :type subnet_id:       string
-        :param port_id:        The port id whose DHCP options are returned
-        :type port_id:         string
-        :returns:              Returns the columns of the DHCP_Options as a
-                               dictionary belonging to the logical switch port
-                               and subnet specified. None is returned if no
-                               DHCP options.
-        """
-
-    @abc.abstractmethod
-    def compose_dhcp_options_commands(self, subnet_id, **columns):
-        """Returns a list of 'Command' objects to add the DHCP options in NB DB
-
-        Checks if there are DHCP_Options rows for the logical switch ports
-        belonging to the @subnet_id and if found adds into the `Command` list.
-
-        @param subnet_id:     The subnet id to which DHCP Options are to be
-                              added
-        @type subnet_id:      string
-        @type columns:        Dictionary of DHCP_Options columns
-        @returns:             List of 'Command' objects returned by
-                              'add_dhcp_options'
+        :param subnet_ids:     The subnet ids whose DHCP options are returned
+        :type subnet_ids:      list of string
+        :returns:              Returns the columns of the DHCP_Options as list
+                               of dictionary. Empty list is returned if no
+                               DHCP_Options matched found.
         """
 
     @abc.abstractmethod
@@ -471,9 +422,208 @@ class API(object):
         :returns: dictionary indexed by name, DB columns as values
         """
 
+    @abc.abstractmethod
+    def get_router_port_options(self, lsp_name):
+        """Get options set for lsp of type router
+
+        :returns: router port options
+        """
+
+    @abc.abstractmethod
+    def add_nat_rule_in_lrouter(self, lrouter, **columns):
+        """Add NAT rule in logical router
+
+
+        :param lrouter:      The unique name of the lrouter
+        :type lrouter:       string
+        :param columns:      Dictionary of nat columns
+                             Supported columns: type, logical_ip, external_ip
+        :type columns:       dictionary
+        :returns:            :class:`Command` with no result
+        """
+
+    @abc.abstractmethod
+    def delete_nat_rule_in_lrouter(self, lrouter, type, logical_ip,
+                                   external_ip, if_exists=True):
+        """Delete NAT rule in logical router
+
+        :param lrouter:      The unique name of the lrouter
+        :type lrouter:       string
+        :param type:         Type of nat. Supported values are 'snat', 'dnat'
+                             and 'dnat_and_snat'
+        :type type:          string
+        :param logical_ip:   IP or network that needs to be natted
+        :type logical_ip:    string
+        :param external_ip:  External IP to be used for nat
+        :type external_ip:   string
+        :param if_exists:    Do not fail if the Logical_Router row does not
+                             exist
+        :type if_exists:     bool
+        :returns:            :class:`Command` with no result
+        """
+
+    @abc.abstractmethod
+    def add_nat_ip_to_lrport_peer_options(self, lport, nat_ip):
+        """Add nat address in peer port of lrouter port
+
+        :param lport:   The unique name of the lswitch port
+        :type lport:    string
+        :param nat_ip:  nat ip to be added
+        :type nat_ip:   string
+        :returns:       :class:`Command` with no result
+        """
+
+    @abc.abstractmethod
+    def delete_nat_ip_from_lrport_peer_options(self, lport, nat_ip):
+        """Delete nat address from peer port of lrouter port
+
+        :param lport:   The unique name of the lswitch port
+        :type lport:    string
+        :param nat_ip:  nat ip to be removed
+        :type nat_ip:   string
+        :returns:       :class:`Command` with no result
+        """
+
+    @abc.abstractmethod
+    def get_lrouter_nat_rules(self, lrouter):
+        """Returns the nat rules of a router
+
+        :param lrouter: The unique name of the router
+        :type lrouter:  string
+        :returns:       A list of nat rules of the router, with each item
+                        as a dict with the keys - 'external_ip', 'logical_ip'
+                        'type' and 'uuid' of the row.
+        """
+
+    @abc.abstractmethod
+    def set_nat_rule_in_lrouter(self, lrouter, nat_rule_uuid, **columns):
+        """Sets the NAT rule fields
+
+        :param lrouter: The unique name of the router to which this the
+                        NAT rule belongs to.
+        :type lrouter:  string
+        :param nat_rule_uuid:  The uuid of the NAT rule row to be updated.
+        :type nat_rule_uuid:   string
+        :type columns:       dictionary
+        :returns:            :class:`Command` with no result
+        """
+
+    @abc.abstractmethod
+    def get_lswitch(self, lswitch_name):
+        """Returns the logical switch
+
+        :param lswitch_name: The unique name of the logical switch
+        :type lswitch_name: string
+        :returns: Returns logical switch or None
+        """
+
+    @abc.abstractmethod
+    def get_ls_and_dns_record(self, lswitch_name):
+        """Returns the logical switch and 'dns' records
+
+        :param lswitch_name: The unique name of the logical switch
+        :type lswitch_name: string
+        :returns: Returns logical switch and dns records as a tuple
+        """
+
+    @abc.abstractmethod
+    def get_floatingip(self, fip_id):
+        """Get a Floating IP by its ID
+
+        :param fip_id: The floating IP id
+        :type fip_id: string
+        :returns: The NAT rule row or None
+        """
+
+    @abc.abstractmethod
+    def get_floatingip_by_ips(self, router_id, logical_ip, external_ip):
+        """Get a Floating IP based on it's logical and external IPs.
+
+        DEPRECATED. In the Rocky release of OpenStack this method can be
+        removed and get_floatingip() should be used instead. This method
+        is a backward compatibility layer for the Pike -> Queens release.
+
+        :param router_id: The ID of the router to which the FIP belongs to.
+        :type lrouter:  string
+        :param logical_ip: The FIP's logical IP address
+        :type logical_ip: string
+        :param external_ip: The FIP's external IP address
+        :type external_ip: string
+        :returns: The NAT rule row or None
+        """
+
+    def check_revision_number(self, name, resource, resource_type,
+                              if_exists=True):
+        """Compare the revision number from Neutron and OVN.
+
+        Check if the revision number in OVN is lower than the one from
+        the Neutron resource, otherwise raise RevisionConflict and abort
+        the transaction.
+
+        :param name:          The unique name of the resource
+        :type name:           string
+        :param resource:      The neutron resource object
+        :type resource:       dictionary
+        :param resource_type: The resource object type
+        :type resource_type:  dictionary
+        :param if_exists:     Do not fail if resource does not exist
+        :type if_exists:      bool
+        :returns:             :class:`Command` with no result
+        :raise:               RevisionConflict if the revision number in
+                              OVN is equal or higher than the neutron object
+        """
+
+    @abc.abstractmethod
+    def get_lswitch_port(self, lsp_name):
+        """Get a Logical Switch Port by its name.
+
+        :param lsp_name: The Logical Switch Port name
+        :type lsp_name: string
+        :returns: The Logical Switch Port row or None
+        """
+
+    @abc.abstractmethod
+    def get_lrouter(self, lrouter_name):
+        """Get a Logical Router by its name
+
+        :param lrouter_name: The name of the logical router
+        :type lrouter_name: string
+        :returns: The Logical_Router row or None
+        """
+
+    @abc.abstractmethod
+    def delete_lrouter_ext_gw(self, lrouter_name):
+        """Delete Logical Router external gateway.
+
+        :param lrouter_name: The name of the logical router
+        :type lrouter_name: string
+        :returns: :class:`Command` with no result
+        """
+
+    @abc.abstractmethod
+    def get_address_set(self, addrset_id, ip_version='ip4'):
+        """Get a Address Set by its ID.
+
+        :param addrset_id: The Address Set ID
+        :type addrset_id: string
+        :param ip_version: Either "ip4" or "ip6". Defaults to "ip4"
+        :type addr_name: string
+        :returns: The Address Set row or None
+        """
+
 
 @six.add_metaclass(abc.ABCMeta)
-class SbAPI(object):
+class SbAPI(api.API):
+
+    @abc.abstractmethod
+    def chassis_exists(self, hostname):
+        """Test if chassis for given hostname exists.
+
+        @param hostname:       The hostname of the chassis
+        @type hostname:        string
+        :returns:              True if the chassis exists, else False.
+        """
+
     @abc.abstractmethod
     def get_chassis_hostname_and_physnets(self):
         """Return a dict contains hostname and physnets mapping.
@@ -482,6 +632,26 @@ class SbAPI(object):
         value. And hostname and physnets are related to the same host.
         """
 
+    def get_gateway_chassis_from_cms_options(self):
+        """Get chassis eligible for external connectivity from CMS options.
+
+        When admin wants to enable router gateway on few chassis,
+        he would set the external_ids as
+
+        ovs-vsctl set open .
+           external_ids:ovn-cms-options="enable-chassis-as-gw"
+        In this function, we parse ovn-cms-options and return these chassis
+        :returns:              List with chassis names.
+        """
+
+    @abc.abstractmethod
+    def get_chassis_and_physnets(self):
+        """Return a dict contains chassis name and physnets mapping.
+
+        Chassis name will be dict key, and a list of physnets will be dict
+        value. And chassis name and physnets are related to the same chassis.
+        """
+
     @abc.abstractmethod
     def get_all_chassis(self, chassis_type=None):
         """Return a list of all chassis which match the compute_type
@@ -489,3 +659,16 @@ class SbAPI(object):
         :param chassis_type:    The type of chassis
         :type chassis_type:     string
         """
+
+    @abc.abstractmethod
+    def get_chassis_data_for_ml2_bind_port(self, hostname):
+        """Return chassis data for ML2 port binding.
+
+        @param hostname:       The hostname of the chassis
+        @type hostname:        string
+        :returns:              Tuple containing the chassis datapath type,
+                               iface types and physical networks for the
+                               OVN bridge mappings.
+        :raises:               RuntimeError exception if an OVN chassis
+                               does not exist.
+        """
